/*
  Copyright 2021 Alain Dargelas

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
*/

/*
 * File:   ParserHarness.cpp
 * Author: alain
 *
 * Created on June 05, 2021, 90:03 AM
 */

#include <Surelog/CommandLine/CommandLineParser.h>
#include <Surelog/Design/FileContent.h>
#include <Surelog/ErrorReporting/ErrorContainer.h>
#include <Surelog/Library/Library.h>
#include <Surelog/SourceCompile/CompilationUnit.h>
#include <Surelog/SourceCompile/Compiler.h>
#include <Surelog/SourceCompile/ParseFile.h>
#include <Surelog/SourceCompile/ParserHarness.h>
#include <Surelog/SourceCompile/SymbolTable.h>

namespace SURELOG {

struct ParserHarness::Holder {
  std::unique_ptr<CompilationUnit> unit;
  std::unique_ptr<SymbolTable> symbols;
  std::unique_ptr<ErrorContainer> errors;
  std::unique_ptr<CommandLineParser> clp;
  std::unique_ptr<Library> lib;
  std::unique_ptr<Compiler> compiler;
  std::unique_ptr<CompileSourceFile> csf;
  std::unique_ptr<ParseFile> pf;
};

std::unique_ptr<FileContent> ParserHarness::parse(std::string_view content) {
  delete m_h;
  m_h = new Holder();

  m_h->unit = std::make_unique<CompilationUnit>(false);
  m_h->symbols = std::make_unique<SymbolTable>();
  m_h->errors = std::make_unique<ErrorContainer>(m_h->symbols.get());
  m_h->clp = std::make_unique<CommandLineParser>(
      m_h->errors.get(), m_h->symbols.get(), false, false);
  m_h->clp->setCacheAllowed(false);
  m_h->lib = std::make_unique<Library>("work", m_h->symbols.get());
  m_h->compiler = std::make_unique<Compiler>(m_h->clp.get(), m_h->errors.get(),
                                             m_h->symbols.get());
  m_h->csf = std::make_unique<CompileSourceFile>(
      BadPathId, m_h->clp.get(), m_h->errors.get(), m_h->compiler.get(),
      m_h->symbols.get(), m_h->unit.get(), m_h->lib.get());
  m_h->pf.reset(
      new ParseFile(content, m_h->csf.get(), m_h->unit.get(), m_h->lib.get()));
  std::unique_ptr<FileContent> file_content_result(
      new FileContent(BadPathId, m_h->lib.get(), m_h->symbols.get(),
                      m_h->errors.get(), nullptr, BadPathId));
  m_h->pf->setFileContent(file_content_result.get());
  if (!m_h->pf->parse()) file_content_result.reset(nullptr);
  return file_content_result;
}

FileContent* ParserHarness::parse(std::string_view content, Compiler* compiler,
                                  PathId fileId) {
  CompilationUnit* unit = new CompilationUnit(false);
  SymbolTable* symbols = compiler->getSymbolTable();
  ErrorContainer* errors = compiler->getErrorContainer();
  CommandLineParser* clp = compiler->getCommandLineParser();
  Library* lib = new Library("work", symbols);
  CompileSourceFile* csf =
      new CompileSourceFile(fileId, clp, errors, compiler, symbols, unit, lib);
  ParseFile* pf = new ParseFile(content, csf, unit, lib);
  FileContent* file_content_result =
      new FileContent(fileId, lib, symbols, errors, nullptr, BadPathId);
  pf->setFileContent(file_content_result);
  if (!pf->parse()) file_content_result = nullptr;
  return file_content_result;
}

ParserHarness::~ParserHarness() { delete m_h; }

}  // namespace SURELOG
